# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import SANITIZE_CC, pack_int, pack_str
from typing import Annotated, Literal, Optional

class MsgClSkinChange(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            skin_part_names: Annotated[list[str], 6] = \
                ['', '', '', '', '', ''],
            use_custom_colors: Annotated[list[bool], 6] = \
                [False, False, False, False, False, False],
            skin_part_colors: Annotated[list[int], 6] = \
                [0, 0, 0, 0, 0, 0]
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'cl_skin_change'
        self.system_message: bool = False
        self.message_id: int = 34
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.7')
        self.header: ChunkHeader = chunk_header

        self.skin_part_names = skin_part_names
        self.use_custom_colors = use_custom_colors
        self.skin_part_colors = skin_part_colors

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'skin_part_names', self.skin_part_names
        yield 'use_custom_colors', self.use_custom_colors
        yield 'skin_part_colors', self.skin_part_colors

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        for i in range(0, 6):
            self.skin_part_names[i] = unpacker.get_str(SANITIZE_CC)
        for i in range(0, 6):
            self.use_custom_colors[i] = unpacker.get_int() == 1
        for i in range(0, 6):
            self.skin_part_colors[i] = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return b''.join([pack_str(x) for x in self.skin_part_names]) + \
            b''.join([pack_int(x) for x in self.use_custom_colors]) + \
            b''.join([pack_int(x) for x in self.skin_part_colors])


#!/usr/bin/env python3

import socket
import time
import argparse
import copy
from signal import signal, SIGINT
from typing import Optional, cast

from twnet_parser.messages7.system.input import MsgInput
from twnet_parser.messages7.system.snap import MsgSnap
from twnet_parser.messages7.system.snap_empty import MsgSnapEmpty
from twnet_parser.messages7.system.snap_single import MsgSnapSingle
from twnet_parser.net_message import NetMessage
from twnet_parser.packer import Unpacker
from twnet_parser.packet import TwPacket7, parse7
from twnet_parser.messages7.control.token import CtrlToken
from twnet_parser.messages7.control.connect import CtrlConnect
from twnet_parser.messages7.control.close import CtrlClose
from twnet_parser.messages7.system.info import MsgInfo
from twnet_parser.messages7.system.map_change import MsgMapChange
from twnet_parser.messages7.system.ready import MsgReady
from twnet_parser.messages7.game.cl_start_info import MsgClStartInfo
from twnet_parser.messages7.system.enter_game import MsgEnterGame

from twnet_parser.constants import NET_MAX_PACKETSIZE, NET_MAX_SEQUENCE
from twnet_parser.session import seq_in_backroom
from twnet_parser.session7 import Session7
from twnet_parser.snap7.character import ObjCharacter
from twnet_parser.snapshot import Snapshot
from twnet_parser.snapshot_storage import SnapshotStorage

example_text = '''example:
 python %(prog)s localhost 8303'''

description_text = '''
 Connects to given server. And keeps the connection alive.
'''

parser = argparse.ArgumentParser(
    prog='snap_printer.py',
    description=description_text,
    epilog=example_text,
    formatter_class=argparse.RawDescriptionHelpFormatter
)
parser.add_argument("host")
parser.add_argument("port", type=int)
args = parser.parse_args()

class TeeworldsClient():
    def __init__(self, host, port):
        self.sock = None
        self.connected = False
        self.connect(host, port)

    def connect(self, host, port):
        self.connected = False
        self.host = host
        self.port = port
        self.dest_srv = (self.host, self.port)

        self.log(f"Connecting to {host}:{port} ...")

        if self.sock:
            self.sock.close()
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.bind(('', 0))

        self.outfile = None
        self.downloaded_chunks = 0
        self.downloaded_bytes = 0
        self.session = Session7()
        self.snap_storage = SnapshotStorage()
        self.input = MsgInput()
        self.map_info: Optional[MsgMapChange] = None
        self.my_token = b'\xff\xaa\xbb\xee'
        self.srv_token = b'\xff\xff\xff\xff'
        self.last_send_time = time.time()

        # TODO: we should be able to set this
        # ctrl_token = CtrlToken(we_are_a_client = True)
        ctrl_token = CtrlToken()
        ctrl_token.response_token = self.my_token
        self.send_msg(ctrl_token)

    def disconnect(self):
        close = CtrlClose()
        self.log("sending disconnect")
        self.send_msg(close)

    def log(self, message):
        print(f"[client] {message}")

    def send_input(self):
        self.send_msg(self.input)

    def send_msg(self, messages):
        if self.sock is None:
            self.log("failed to send message! No socket!")
            return

        if not isinstance(messages, list):
            messages = [messages]
        packet = TwPacket7()
        packet.header.token = self.srv_token
        for msg in messages:
            if hasattr(msg, 'header'):
                if msg.header.flags.vital:
                    self.session.sequence += 1
                msg.header.seq = self.session.sequence
            packet.messages.append(msg)
        self.last_send_time = time.time()
        packet.header.ack = self.session.ack
        self.sock.sendto(packet.pack(), self.dest_srv)

    def on_snapshot(self, snapshot: Snapshot):
        for item in snapshot.items:
            if item.item_name != 'obj.character':
                continue
            char: ObjCharacter = cast(ObjCharacter, item)
            self.log(f"cid={char.id} at x={char.x / 32} y={char.y / 32}")

    def tick(self):
        if self.sock is None:
            self.log("You have to call connect() before you can tick()")
            return
        data, _ = self.sock.recvfrom(NET_MAX_PACKETSIZE)
        packet = parse7(data)

        for msg in packet.messages:
            if hasattr(msg, 'header'):
                msg = cast(NetMessage, msg)
                if msg.header.flags.vital:
                    if msg.header.seq == (self.session.ack + 1) % NET_MAX_SEQUENCE:
                        # in sequence
                        self.session.ack = (self.session.ack + 1) % NET_MAX_SEQUENCE
                    else:
                        # old packet that we already got
                        if seq_in_backroom(msg.header.seq, self.session.ack):
                            continue
                        # out of sequence, request resend
                        self.log("need to request resend, but that is not implemented yet!")

            if msg.message_name.startswith('snap_') and not self.connected:
                self.connected = True
                self.log("connected!")

            if msg.message_name == 'token':
                msg = cast(CtrlToken, msg)
                self.srv_token = msg.response_token

                ctrl_connect = CtrlConnect()
                ctrl_connect.response_token = self.my_token
                self.send_msg(ctrl_connect)
            elif msg.message_name == 'accept':
                info = MsgInfo()
                info.header.flags.vital = True
                self.send_msg(info)
            elif msg.message_name == 'map_change' or msg.message_name == 'map_data':
                ready = MsgReady()
                ready.header.flags.vital = True
                self.send_msg(ready)
            elif msg.message_name == 'con_ready':
                self.log("sending info")
                startinfo = MsgClStartInfo()
                startinfo.header.flags.vital = True
                self.send_msg(startinfo)
                enter_game = MsgEnterGame()
                enter_game.header.flags.vital = True
                self.send_msg(enter_game)
            elif msg.message_name == 'close':
                msg = cast(CtrlClose, msg)
                self.log(f"disconnected reason='{msg.reason}'")
                exit(1)
            elif msg.message_name == 'snap_single':
                msg = cast(MsgSnapSingle, msg)
                delta_tick = msg.tick - msg.delta_tick
                self.log(f"got snap single with tick={msg.tick} delta_tick={delta_tick}")

                # # TODO: this comparison looks wrong to me but its the same in teeworlds-go
                # if self.snap_storage.current_recv_tick < msg.tick:
                #     self.log(f"dropping snap with too old game tick msg.tick={msg.tick} current_recv_tick={self.snap_storage.current_recv_tick}")
                #     continue

                if msg.tick != self.snap_storage.current_recv_tick:
                    self.snap_storage.snapshot_parts = 0
                    self.snap_storage.current_recv_tick = msg.tick

                # empty snap if delta is -1
                prev_snap = Snapshot(version='0.7')

                if delta_tick != -1:
                    prev_snap = self.snap_storage.get(delta_tick)

                if not prev_snap:
                    raise ValueError(f"delta snapshot with tick {delta_tick} not found")

                unpacker = Unpacker(msg.data)
                new_full_snap = prev_snap.unpack_delta(unpacker)
                self.snap_storage.add(msg.tick, new_full_snap)
                self.snap_storage.purge_until(delta_tick)

                self.on_snapshot(new_full_snap)

                self.input.ack_snapshot = msg.tick
                self.input.intended_tick = self.snap_storage.newest_tick
                self.snap_storage.current_recv_tick = msg.tick

                self.log(f"intended tick {self.input.intended_tick}")

                # TODO: should we send input here? teeworlds-go/protocol does it

            elif msg.message_name == 'snap_empty':
                msg = cast(MsgSnapEmpty, msg)
                delta_tick = msg.tick - msg.delta_tick
                self.log(f"got snap empty with tick={msg.tick} delta_tick={delta_tick}")

                # # TODO: this comparison looks wrong to me but its the same in teeworlds-go
                # if self.snap_storage.current_recv_tick < msg.tick:
                #     self.log(f"dropping snap with too old game tick msg.tick={msg.tick} current_recv_tick={self.snap_storage.current_recv_tick}")
                #     continue

                if msg.tick != self.snap_storage.current_recv_tick:
                    self.snap_storage.snapshot_parts = 0
                    self.snap_storage.current_recv_tick = msg.tick

                # empty snap if delta is -1
                prev_snap = Snapshot(version='0.7')

                if delta_tick != -1:
                    prev_snap = self.snap_storage.get(delta_tick)

                if not prev_snap:
                    raise ValueError(f"delta snapshot with tick {delta_tick} not found")

                # TODO: the deepcopy can probably be avoided
                #       if we correctly drop or purge unused snaps
                self.snap_storage.add(msg.tick, copy.deepcopy(prev_snap))
                self.snap_storage.purge_until(delta_tick)

                self.on_snapshot(prev_snap)

                self.input.ack_snapshot = msg.tick
                self.input.intended_tick = self.snap_storage.newest_tick
                self.snap_storage.current_recv_tick = msg.tick
            elif msg.message_name == 'snap':
                msg = cast(MsgSnap, msg)
                delta_tick = msg.tick - msg.delta_tick
                self.log(f"got snap part {msg.part}/{msg.num_parts} tick={msg.tick} delta_tick={delta_tick}")

                # TODO: this comparison looks wrong to me but its the same in teeworlds-go
                if self.snap_storage.current_recv_tick < msg.tick:
                    self.log(f"dropping snap with too old game tick msg.tick={msg.tick} current_recv_tick={self.snap_storage.current_recv_tick}")
                    continue

                if msg.tick != self.snap_storage.current_recv_tick:
                    self.snap_storage.snapshot_parts = 0
                    self.snap_storage.current_recv_tick = msg.tick

                self.snap_storage.add_incoming_data(msg.part, msg.num_parts, msg.data)

                self.snap_storage.snapshot_parts |= 1 << msg.part

                if msg.part != msg.num_parts -1:
                    # TODO: remove this if it works
                    self.log(f"storing partial snap part={msg.part} num_parts={msg.num_parts}")
                    continue

                if self.snap_storage.snapshot_parts != msg.num_parts - 1:
                    self.log(f"got last part but mussing previous parts part={msg.part} num_parts={msg.num_parts}")
                    continue

                self.snap_storage.snapshot_parts = 0

                # empty snap if delta is -1
                prev_snap = Snapshot(version='0.7')

                if delta_tick != -1:
                    prev_snap = self.snap_storage.get(delta_tick)

                if not prev_snap:
                    raise ValueError(f"delta snapshot with tick {delta_tick} not found")

                unpacker = Unpacker(self.snap_storage.incoming_data())
                new_full_snap = prev_snap.unpack_delta(unpacker)
                self.snap_storage.add(msg.tick, new_full_snap)
                self.snap_storage.purge_until(delta_tick)

                self.on_snapshot(new_full_snap)

                self.input.ack_snapshot = msg.tick
                self.input.intended_tick = self.snap_storage.newest_tick
                self.snap_storage.current_recv_tick = msg.tick

        if (time.time() - self.last_send_time) > 1:
            self.send_input()

client = TeeworldsClient(host = args.host, port = args.port)

def handler(signal_received, _):
    global client
    print(f"got signal: {signal_received}")
    print('SIGINT or CTRL-C detected. Exiting gracefully')
    client.disconnect()
    exit(0)

signal(SIGINT, handler)

while True:
    client.tick()




from typing import Dict, Optional
from twnet_parser.snapshot import Snapshot

class Holder:
    def __init__(self, tick: int, snapshot: Snapshot):
        self.snapshot: Snapshot = snapshot
        self.tick: int = tick

class SnapshotStorage:
    # passed to methods
    EMPTY_SNAP_TICK = -1

    # returned by methods
    UNINITIALIZED_TICK = -1

    def __init__(self) -> None:
        self.holder: Dict[int, Holder] = {}

        # oldest tick still in the holder
        # not the oldest tick we ever received
        self.oldest_tick = self.UNINITIALIZED_TICK

        # newest tick in the holder
        self.newest_tick = self.UNINITIALIZED_TICK

        # the tick we are currently collecting parts for
        self.current_recv_tick: int = 0

        # received parts for the current tick
        # as a bit field
        # to check if we received all previous parts
        # when we get the last part number
        self.snapshot_parts: int = 0

        self.__multi_incoming_data: bytes

    def next_tick(self, tick: int) -> int:
        for k in sorted(list(self.holder.keys())):
            if k > tick:
                return k
        return self.UNINITIALIZED_TICK

    def previous_tick(self, tick: int) -> int:
        for k in reversed(sorted(list(self.holder.keys()))):
            if k < tick:
                return k
        return self.UNINITIALIZED_TICK

    def purge_until(self, tick: int) -> None:
        for k in list(self.holder.keys()):
            if k < tick:
                del self.holder[k]

        if self.oldest_tick != self.UNINITIALIZED_TICK:
            found = self.oldest_tick in self.holder
            if not found:
                self.oldest_tick = self.next_tick(self.oldest_tick)

        if self.newest_tick != self.UNINITIALIZED_TICK:
            found = self.oldest_tick in self.holder
            if not found:
                self.newest_tick = self.next_tick(self.newest_tick)

    def add_incoming_data(self, part: int, num_parts: int, data: bytes) -> None:
        if part == 0:
            # reset length if we get a new snapshot
            self.__multi_incoming_data = b''

        if part != num_parts - 1:
            if len(data) != Snapshot.MAX_PACK_SIZE:
                raise ValueError( \
                                 "Error: incomplete part that is not the last expected part:" \
                                 f"part={part} num_parts={num_parts} " \
                                 f"expected_size={Snapshot.MAX_PACK_SIZE} got_size={len(data)}")
        if len(self.__multi_incoming_data) + len(data) > Snapshot.MAX_SIZE:
            raise ValueError("Error: " \
                f"reached the maximum amount of snapshot data: {Snapshot.MAX_SIZE}")

        self.__multi_incoming_data += data

    def incoming_data(self) -> bytes:
        return self.__multi_incoming_data

    def get(self, tick: int) -> Optional[Snapshot]:
        if tick == self.EMPTY_SNAP_TICK:
            # TODO: can we return an empty snap here?
            #       not really because we do not know if 0.7 or 0.6
            return None
        if tick < 0:
            raise ValueError(f"Error: negative ticks not supported! Tried to get tick {tick}")
        if tick not in self.holder:
            return None

        return self.holder[tick].snapshot

    def add(self, tick: int, snapshot: Snapshot) -> None:
        if tick < 0:
            raise ValueError(f"Error: negative ticks not supported! Tried to add tick {tick}")
        new_snap_id = id(snapshot)
        for holder in list(self.holder.values()):
            if id(holder.snapshot) == new_snap_id:
                raise ValueError(
                    f"Error: snapshot with object id {new_snap_id} already in the storage." \
                        "Use copy.deepcopy() to avoid bugs.")

        if self.oldest_tick == self.UNINITIALIZED_TICK or tick < self.oldest_tick:
            self.oldest_tick = tick
        if self.newest_tick == self.UNINITIALIZED_TICK or tick > self.newest_tick:
            self.newest_tick = tick
        self.holder[tick] = Holder(tick, snapshot)



# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import SANITIZE_CC, pack_int, pack_str
from typing import Literal, Optional

class MsgClStartInfo(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            name: str = 'default',
            clan: str = 'default',
            country: int = 0,
            skin: str = 'default',
            use_custom_color: bool = False,
            color_body: int = 0,
            color_feet: int = 0
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'cl_start_info'
        self.system_message: bool = False
        self.message_id: int = 20
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.6')
        self.header: ChunkHeader = chunk_header

        self.name: str = name
        self.clan: str = clan
        self.country: int = country
        self.skin: str = skin
        self.use_custom_color: bool = use_custom_color
        self.color_body: int = color_body
        self.color_feet: int = color_feet

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'name', self.name
        yield 'clan', self.clan
        yield 'country', self.country
        yield 'skin', self.skin
        yield 'use_custom_color', self.use_custom_color
        yield 'color_body', self.color_body
        yield 'color_feet', self.color_feet

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.name = unpacker.get_str(SANITIZE_CC)
        self.clan = unpacker.get_str(SANITIZE_CC)
        self.country = unpacker.get_int()
        self.skin = unpacker.get_str(SANITIZE_CC)
        self.use_custom_color = unpacker.get_int() == 1
        self.color_body = unpacker.get_int()
        self.color_feet = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_str(self.name) + \
            pack_str(self.clan) + \
            pack_int(self.country) + \
            pack_str(self.skin) + \
            pack_int(self.use_custom_color) + \
            pack_int(self.color_body) + \
            pack_int(self.color_feet)


from typing import Optional

from twnet_parser.packer import Unpacker
from twnet_parser.snap.unknown import ObjUnknown
from twnet_parser.snap_item import SnapItem

import twnet_parser.obj6

from twnet_parser.snap6.player_input import ObjPlayerInput
from twnet_parser.snap6.projectile import ObjProjectile
from twnet_parser.snap6.laser import ObjLaser
from twnet_parser.snap6.pickup import ObjPickup
from twnet_parser.snap6.flag import ObjFlag
from twnet_parser.snap6.game_info import ObjGameInfo
from twnet_parser.snap6.game_data import ObjGameData
from twnet_parser.snap6.character_core import ObjCharacterCore
from twnet_parser.snap6.character import ObjCharacter
from twnet_parser.snap6.player_info import ObjPlayerInfo
from twnet_parser.snap6.client_info import ObjClientInfo
from twnet_parser.snap6.spectator_info import ObjSpectatorInfo
from twnet_parser.snap6.common import ObjCommon
from twnet_parser.snap6.explosion import ObjExplosion
from twnet_parser.snap6.spawn import ObjSpawn
from twnet_parser.snap6.hammer_hit import ObjHammerHit
from twnet_parser.snap6.death import ObjDeath
from twnet_parser.snap6.sound_global import ObjSoundGlobal
from twnet_parser.snap6.sound_world import ObjSoundWorld
from twnet_parser.snap6.damage_ind import ObjDamageInd

def match_item(unpacker: Unpacker) -> Optional[SnapItem]:
    item_type = unpacker.get_int()
    item: SnapItem
    if item_type == twnet_parser.obj6.PLAYER_INPUT:
        item = ObjPlayerInput()
    elif item_type == twnet_parser.obj6.PROJECTILE:
        item = ObjProjectile()
    elif item_type == twnet_parser.obj6.LASER:
        item = ObjLaser()
    elif item_type == twnet_parser.obj6.PICKUP:
        item = ObjPickup()
    elif item_type == twnet_parser.obj6.FLAG:
        item = ObjFlag()
    elif item_type == twnet_parser.obj6.GAME_INFO:
        item = ObjGameInfo()
    elif item_type == twnet_parser.obj6.GAME_DATA:
        item = ObjGameData()
    elif item_type == twnet_parser.obj6.CHARACTER_CORE:
        item = ObjCharacterCore()
    elif item_type == twnet_parser.obj6.CHARACTER:
        item = ObjCharacter()
    elif item_type == twnet_parser.obj6.PLAYER_INFO:
        item = ObjPlayerInfo()
    elif item_type == twnet_parser.obj6.CLIENT_INFO:
        item = ObjClientInfo()
    elif item_type == twnet_parser.obj6.SPECTATOR_INFO:
        item = ObjSpectatorInfo()
    elif item_type == twnet_parser.obj6.COMMON:
        item = ObjCommon()
    elif item_type == twnet_parser.obj6.EXPLOSION:
        item = ObjExplosion()
    elif item_type == twnet_parser.obj6.SPAWN:
        item = ObjSpawn()
    elif item_type == twnet_parser.obj6.HAMMERHIT:
        item = ObjHammerHit()
    elif item_type == twnet_parser.obj6.DEATH:
        item = ObjDeath()
    elif item_type == twnet_parser.obj6.SOUND_GLOBAL:
        item = ObjSoundGlobal()
    elif item_type == twnet_parser.obj6.SOUND_WORLD:
        item = ObjSoundWorld()
    elif item_type == twnet_parser.obj6.DAMAGE_INDICATOR:
        item = ObjDamageInd()
    else:
        item = ObjUnknown()
        item.type_id = item_type
    return item




# generated by scripts/generate_messages.py

from twnet_parser.pretty_print import PrettyPrint
from twnet_parser.packer import Unpacker
from twnet_parser.chunk_header import ChunkHeader
from twnet_parser.packer import pack_int
from typing import Literal, Optional

class MsgSvServerSettings(PrettyPrint):
    def __init__(
            self,
            chunk_header: Optional[ChunkHeader] = None,
            kick_vote: bool = False,
            kick_min: int = 0,
            spec_vote: bool = False,
            team_lock: bool = False,
            team_balance: bool = False,
            player_slots: int = 0
    ) -> None:
        self.message_type: Literal['system', 'game'] = 'game'
        self.message_name: str = 'sv_server_settings'
        self.system_message: bool = False
        self.message_id: int = 17
        if not chunk_header:
            chunk_header = ChunkHeader(version = '0.7')
        self.header: ChunkHeader = chunk_header

        self.kick_vote: bool = kick_vote
        self.kick_min: int = kick_min
        self.spec_vote: bool = spec_vote
        self.team_lock: bool = team_lock
        self.team_balance: bool = team_balance
        self.player_slots: int = player_slots

    def __iter__(self):
        yield 'message_type', self.message_type
        yield 'message_name', self.message_name
        yield 'system_message', self.system_message
        yield 'message_id', self.message_id
        yield 'header', dict(self.header)

        yield 'kick_vote', self.kick_vote
        yield 'kick_min', self.kick_min
        yield 'spec_vote', self.spec_vote
        yield 'team_lock', self.team_lock
        yield 'team_balance', self.team_balance
        yield 'player_slots', self.player_slots

    # first byte of data
    # has to be the first byte of the message payload
    # NOT the chunk header and NOT the message id
    def unpack(self, data: bytes) -> bool:
        unpacker = Unpacker(data)
        self.kick_vote = unpacker.get_int() == 1
        self.kick_min = unpacker.get_int()
        self.spec_vote = unpacker.get_int() == 1
        self.team_lock = unpacker.get_int() == 1
        self.team_balance = unpacker.get_int() == 1
        self.player_slots = unpacker.get_int()
        return True

    def pack(self) -> bytes:
        return pack_int(self.kick_vote) + \
            pack_int(self.kick_min) + \
            pack_int(self.spec_vote) + \
            pack_int(self.team_lock) + \
            pack_int(self.team_balance) + \
            pack_int(self.player_slots)


# pylint: disable=duplicate-code
PLAYER_INPUT = 1
PROJECTILE = 2
LASER = 3
PICKUP = 4
FLAG = 5
GAME_INFO = 6
GAME_DATA = 7
CHARACTER_CORE = 8
CHARACTER = 9
PLAYER_INFO = 10
CLIENT_INFO = 11
SPECTATOR_INFO = 12
COMMON = 13
EXPLOSION = 14
SPAWN = 15
HAMMERHIT = 16
DEATH = 17
SOUND_GLOBAL = 18
SOUND_WORLD = 19
DAMAGE_INDICATOR = 20



# pylint: disable=duplicate-code
from enum import Enum

class Pickup(Enum):
    HEALTH = 0
    ARMOR = 1
    GRENADE = 2
    SHOTGUN = 3
    LASER = 4
    NINJA = 5
    GUN = 6
    HAMMER = 7

class Emote(Enum):
    NORMAL = 0
    PAIN = 1
    HAPPY = 2
    SURPRISE = 3
    ANGRY = 4
    BLINK = 5

class Emoticon(Enum):
    OOP = 0
    EXCLAMATION = 1
    HEARTS = 2
    DROP = 3
    DOTDOT = 4
    MUSIC = 5
    SORRY = 6
    GHOST = 7
    SUSHI = 8
    SPLATTEE = 9
    DEVILTEE = 10
    ZOMG = 11
    ZZZ = 12
    WTF = 13
    EYES = 14
    QUESTION = 15

class Vote(Enum):
    UNKNOWN = 0
    START_OP = 1
    START_KICK = 2
    START_SPEC = 3
    END_ABORT = 4
    END_PASS = 5
    END_FAIL = 6

class Chat(Enum):
    NONE = 0
    ALL = 1
    TEAM = 2
    WHISPER = 3

class Gamemsg(Enum):
    TEAM_SWAP = 0
    SPEC_INVALIDID = 1
    TEAM_SHUFFLE = 2
    TEAM_BALANCE = 3
    CTF_DROP = 4
    CTF_RETURN = 5
    TEAM_ALL = 6
    TEAM_BALANCE_VICTIM = 7
    CTF_GRAB = 8
    CTF_CAPTURE = 9
    GAME_PAUSED = 10

class Weapon(Enum):
    HAMMER = 0
    PISTOL = 1
    SHOTGUN = 2
    GRENADE = 3
    RIFLE = 4
    NINJA = 5

class Team(Enum):
    SPECTATORS = -1
    RED = 0
    BLUE = 1

class Sound(Enum):
    GUN_FIRE = 0
    SHOTGUN_FIRE = 1
    GRENADE_FIRE = 2
    HAMMER_FIRE = 3
    HAMMER_HIT = 4
    NINJA_FIRE = 5
    GRENADE_EXPLODE = 6
    NINJA_HIT = 7
    RIFLE_FIRE = 8
    RIFLE_BOUNCE = 9
    WEAPON_SWITCH = 10
    PLAYER_PAIN_SHORT = 11
    PLAYER_PAIN_LONG = 12
    BODY_LAND = 13
    PLAYER_AIRJUMP = 14
    PLAYER_JUMP = 15
    PLAYER_DIE = 16
    PLAYER_SPAWN = 17
    PLAYER_SKID = 18
    TEE_CRY = 19
    HOOK_LOOP = 20
    HOOK_ATTACH_GROUND = 21
    HOOK_ATTACH_PLAYER = 22
    HOOK_NOATTACH = 23
    PICKUP_HEALTH = 24
    PICKUP_ARMOR = 25
    PICKUP_GRENADE = 26
    PICKUP_SHOTGUN = 27
    PICKUP_NINJA = 28
    WEAPON_SPAWN = 29
    WEAPON_NOAMMO = 30
    HIT = 31
    CHAT_SERVER = 32
    CHAT_CLIENT = 33
    CHAT_HIGHLIGHT = 34
    CTF_DROP = 35
    CTF_RETURN = 36
    CTF_GRAB_PL = 37
    CTF_GRAB_EN = 38
    CTF_CAPTURE = 39
    MENU = 40

class Spec(Enum):
    FREEVIEW = 0
    PLAYER = 1
    FLAGRED = 2
    FLAGBLUE = 3

class Skinpart(Enum):
    BODY = 0
    MARKING = 1
    DECORATION = 2
    HANDS = 3
    FEET = 4
    EYES = 5
